import requests
import aiohttp
from GramDB.method import *
from GramDB.helper import EfficientDictQuery
from GramDB.exception import *
import asyncio

class GramDB:
    def __init__(self, db_url: str):
        self.db_url = db_url
        self.session = None
        self.token = None
        self.url = None
        self.CACHE_TABLE = None
        self.CACHE_DATA = None
        self.db = None
        self.initialize()

    def initialize(self):
        self.session = aiohttp.ClientSession()
        self.authenticate()

    def authenticate(self):
        try:
            response = requests.get(self.db_url)
            if response.status_code == 400:
                raise ValueError("Authentication failed: Invalid credentials or URL.")
            elif response.status_code != 200:
                raise ValueError(f"Authentication failed: Unexpected status code {response.status_code}")
                                                                                  
            self.auth = response.json()
            self.token = self.auth['client_id']
            self.url = self.auth['url']
            self.import_cache()
        
        except Exception as e:
            raise ConnectionError(f"Network error during authentication: {e}")

    def import_cache(self):
        try:
            result, data = extract_func(self.url, self.token)
            if result:
                self.CACHE_TABLE = data
            else:
                raise ValidationError("Authentication failed: token expired or outdated!")
            
            self.CACHE_DATA = {}
            all_ids = []
            for tablename, table in self.CACHE_TABLE.items():
                if tablename=="info_gramdb":
                    pass
                else:
                    all_ids = all_ids + table
                
            result, all_rows = fetchall_func(self.url, self.token, all_ids)
            if result:
                for row in all_rows["data"]:
                    self.CACHE_DATA[row['_m_id']] = row
            else:
                raise GramDBError(f"Failed to fetch all data: {all_rows}")
                
            self.db = EfficientDictQuery(self.CACHE_DATA)
        except Exception as e:
            raise GramDBError(f"Error importing cache: {e}")

    async def background_create(self, table_name, _m_id):
        try:
            result, old_data = extract_func(self.url, self.token)
            old_data[table_name] = [_m_id]
            result2, response = await git_func(self.session, self.url, self.token, old_data)
        except Exception as e:
            raise GramDBError(f"Error in background create: {e}")

    async def create(self, table_name: str, schema):
        try:
            sample_record = {field: "test" for field in schema}
            sample_record['_id'] = "sample1928"
            result, mdata = await insert_func(self.session, self.url, self.token, sample_record, table_name)
            if result:
                _m_id = mdata["data_id"]
                sample_record['_m_id'] = _m_id
                del sample_record["_table_"]
                await self.db.create(table_name, schema, sample_record, _m_id)
                asyncio.create_task(self.background_create(table_name, _m_id))
            else:
                raise GramDBError(f"Failed to create record in table {table_name}")
        except Exception as e:
            raise GramDBError(f"Error creating record: {e}")

    async def fetch(self, table_name: str, query: dict):
        try:
            result = await self.db.fetch(table_name, query)
            print(result)
            #if not result:
            #    raise NotFoundError(f"Record not found in table {table_name} for query {query}")
            if len(result) == 0:
                return None
            elif len(result) == 1:
                return result[0]
            else:
                return result
        except Exception as e:
            raise GramDBError(f"Error fetching data: {e}")

    async def fetch_all(self):
        try:
            return await self.db.fetch_all()
        except Exception as e:
            raise GramDBError(f"Error fetching all data: {e}")

    async def background_insert(self, table_name, _m_id):
        try:
            result, old_data = extract_func(self.url, self.token)
            new_data = old_data[table_name]
            new_data.append(_m_id)
            result2, response = await git_func(self.session, self.url, self.token, old_data)
        except Exception as e:
            raise GramDBError(f"Error in background insert: {e}")

    async def insert(self, table_name: str, record: dict):
        try:
            if '_id' not in record:
                record['_id'] = await self.db._generate_random_id()

            result, mdata = await insert_func(self.session, self.url, self.token, record, table_name)
            if result:
                _m_id = mdata["data_id"]
                record['_m_id'] = _m_id
                del record['_table_']
                await self.db.insert(table_name, record, _m_id=_m_id)
                asyncio.create_task(self.background_insert(table_name, _m_id))
            else:
                raise GramDBError(f"Failed to insert record in table {table_name}")
        except Exception as e:
            raise GramDBError(f"Error inserting record: {e}")

    async def background_delete(self, table_name, _m_id):
        try:
            result, old_data = extract_func(self.url, self.token)
            new_data = old_data[table_name]          
            new_data.remove(int(_m_id))
            result2, response = await git_func(self.session, self.url, self.token, old_data)
        except Exception as e:
            raise GramDBError(f"Error in background delete: {e}")

    async def delete(self, table_name: str, query: dict):
        try:
            _m_id = await self.db.delete(table_name, query)
            asyncio.create_task(self.background_delete(table_name, _m_id))
        except Exception as e:
            raise GramDBError(f"Error deleting record: {e}")

    async def background_update(self, table_name, update_query, _m_id):
        try:
            await asyncio.sleep(8)
            records = await self.db.fetch(table_name, update_query)       
            result, mdata = await update_func(self.session, self.url, self.token, _m_id, records[0], table_name)
        except Exception as e:
            raise GramDBError(f"Error in background update: {e}")

    async def update(self, table_name: str, query: dict, update_query: dict):
        try:
            _m_id = await self.db.update(table_name, query, update_query)
            asyncio.create_task(self.background_update(table_name, update_query, _m_id))
        except Exception as e:
            raise GramDBError(f"Error updating record: {e}")

    async def background_delete_table(self, table_name):
        try:
            result, old_data = extract_func(self.url, self.token)
            del old_data[table_name]          
            result2, response = await git_func(self.session, self.url, self.token, old_data)
        except Exception as e:
            raise GramDBError(f"Error in background delete table: {e}")
            
    async def delete_table(self, table_name: str):
        try:
            await self.db.delete_table(table_name)
            asyncio.create_task(self.background_delete_table(table_name))
        except Exception as e:
            raise GramDBError(f"Error deleting table {table_name}: {e}")

    async def close_func(self):
        await self.session.close()
        
    def close(self):
        loop = asyncio.get_event_loop()
        if loop.is_running():
            loop.create_task(self.close_func())
        else:
            asyncio.run(self.close_func())



class GramDBError(Exception):
    pass

class ConnectionError(GramDBError):
    pass

class NotFoundError(GramDBError):
    pass

class DuplicateTableError(GramDBError):
    pass

class ValidationError(GramDBError):
    pass
    




# methods
import requests
import json
import copy


async def fetch_func(session, base_url, token, data_id):
    url = base_url + "/fetch"
    headers = {
        "Content-Type": "application/json",
        "Authorization": token
    }
    payload = {
        "data_id": data_id
    }
    response = await session.post(url, headers=headers, json=payload)
    if response.status == 200:
        result = await response.json()
        return True, result
    else:
        return False, response.status


async def insert_func(session, base_url, token, data, table_name):
    url = base_url + "/insert"
    headers = {
        "Content-Type": "application/json",
        "Authorization": token
    }
    data["_table_"] = table_name
    dataa = json.dumps(data)
    payload = {
        "data": str(dataa)
    }
    async with session.post(url, headers=headers, json=payload) as response:
        if response.status == 201:
            result = await response.json()
            return True, result
        else:
            return False, response.status


async def delete_func(session, base_url, token, data_id):
    url = base_url + "/delete"
    headers = {
        "Content-Type": "application/json",
        "Authorization": token
    }
    payload = {
        "data_id": data_id
    }
    response = await session.post(url, headers=headers, json=payload)
    if response.status == 200:
        result = await response.json()
        return True, result
    else:
        return False, response.status


async def update_func(session, base_url, token, data_id, data, table_name):
    dataa = copy.deepcopy(data)
    dataa["_table_"] = table_name
    del dataa["_m_id"]
    dataa = json.dumps(dataa)
    
    url = base_url + "/update"
    headers = {
        "Content-Type": "application/json",
        "Authorization": token
    }
    payload = {
        "data_id": data_id,
        "data": dataa
    }
    response = await session.post(url, headers=headers, json=payload)
    if response.status == 200:
        result = await response.json()
        return True, result
    else:
        return False, response.status


async def git_func(session, base_url, token, data):
    url = base_url + "/git"
    headers = {
        "Content-Type": "application/json",
        "Authorization": token
    }
    payload = {
        "data" : data
    }
    response = await session.post(url, headers=headers, json=payload)
    if response.status == 200:
        result = await response.json()
        return True, result
    else:
        return False, response.status


def extract_func(base_url, token):
    url = base_url + f"/extract?token={token}"
    headers = {
        "Content-Type": "application/json",
    }
    response = requests.get(url, headers=headers)
    if response.status_code == 200:
        result = response.json()
        return True, result
    else:
        return False, response.status_code
    

def fetchall_func(base_url, token, data_ids):
    url = base_url + "/fetchall"
    headers = {
        "Content-Type": "application/json",
        "Authorization": token
    }
    payload = {
        "data_id": data_ids
    }
    response = requests.post(url, headers=headers, json=payload)
    if response.status_code == 200:
        result = response.json()
        return True, result
    else:
        return False, response.status_code
        



from collections import defaultdict
import random
import string

class EfficientDictQuery:
    def __init__(self, data):
        self.data = self._structure_data(data)
        self.indexes = defaultdict(lambda: defaultdict(list))
        self.schemas = {}
        self.create_all_indexes()
        self.create_all_schemas()

    def _structure_data(self, data):
        structured_data = defaultdict(dict)
        for record in data.values():
            table = record['_table_']
            primary_key = record['_id']
            structured_record = {k: v for k, v in record.items() if k not in ['_table_']}
            structured_data[table][str(primary_key)] = structured_record
        return structured_data

    def create_all_indexes(self):
        if not self.data:
            return

        fields = set()
        for table in self.data.values():
            for item in table.values():
                flattened_item = self._flatten_dict(item)
                fields.update(flattened_item.keys())
        
        for field in fields:
            self.create_index(field)

    def create_index(self, field):
        index = defaultdict(list)
        for table_name, table in self.data.items():
            for key, item in table.items():
                flattened_item = self._flatten_dict(item)
                if field in flattened_item:
                    index[flattened_item[field]].append((table_name, key))
        self.indexes[field] = index

    def _flatten_dict(self, d, parent_key='', sep='.'):
        items = []
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict):
                items.extend(self._flatten_dict(v, new_key, sep=sep).items())
            elif isinstance(v, list):
                for i, item in enumerate(v):
                    items.extend(self._flatten_dict({str(i): item}, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)

    def create_all_schemas(self):
        for table_name, records in self.data.items():
            if table_name not in self.schemas:
                schema = set()
                for record in records.values():
                    schema.update(record.keys())
                self.schemas[table_name] = tuple(schema)
    
    async def fetch(self, table, query):
        results = []

        for record_id, record in self.data.get(table, {}).items():
            if all(record.get(key) == value for key, value in query.items()):
                results.append(record)

        return results

    async def _update_index_for_record(self, table, record, record_id, operation='add'):
        flattened_record = self._flatten_dict(record)
        for field, value in flattened_record.items():
            if operation == 'add':
                self.indexes[field][value].append((table, record_id))
            elif operation == 'remove':
                if (table, record_id) in self.indexes[field][value]:
                    self.indexes[field][value].remove((table, record_id))
                    if not self.indexes[field][value]:
                        del self.indexes[field][value]

    async def _validate_record(self, table, record):
        if table not in self.schemas:
            raise ValueError(f"Table '{table}' does not exist.")

        schema = self.schemas[table]
        for field in schema:
            if field not in record:
                raise ValueError(f"Missing required field '{field}' in record for table '{table}'.")

        for field in record:
            if field not in schema:
                raise ValueError(f"Field '{field}' is not allowed in schema for table '{table}'.")

    async def create(self, table, schema, sample_record, _m_id):
        if table in self.data:
            raise ValueError(f"Table '{table}' already exists.")

        schema = set(schema)
        schema.update(["_id", "_m_id"])
        self.schemas[table] = tuple(schema)

        self.data[table] = {"sample1928": sample_record}
        await self._update_index_for_record(table, sample_record, "sample1928", operation='add')

    async def _generate_random_id(self):
        return ''.join(random.choices(string.ascii_letters + string.digits, k=10))
                               
    async def insert(self, table, record, **kwargs):
        _m_id = kwargs.get('_m_id')
        if not _m_id:
            raise ValueError("Record must contain '_m_id' as a keyword argument.")

        if table not in self.data:
            raise ValueError(f"Invalid table name '{table}'. Table does not exist.")
            
        _id = str(record['_id'])
        record['_m_id'] = _m_id

        await self._validate_record(table, record)

        if _id in self.data[table]:
            raise ValueError(f"Record with _id '{_id}' already exists in table '{table}'.")

        self.data[table][_id] = record
        await self._update_index_for_record(table, record, _id, operation='add')

    async def update(self, table, query, update_fields):
        if table not in self.data:
            raise ValueError(f"Table '{table}' does not exist.")

        records_to_update = [
            (record_id, record) for record_id, record in self.data[table].items()
            if all(record.get(key) == value for key, value in query.items())
        ]

        if not records_to_update:
            raise ValueError(f"No records found matching query: {query}")

        record_id, old_record = records_to_update[0]
        _m_id = old_record["_m_id"]
        combined_record = {**old_record, **update_fields}
        await self._validate_record(table, combined_record)
        await self._update_index_for_record(table, old_record, record_id, operation='remove')

        self.data[table][record_id].update(update_fields)
        await self._update_index_for_record(table, self.data[table][record_id], record_id, operation='add')
        return _m_id
    
    async def delete(self, table, query):
        if table not in self.data:
            raise ValueError(f"Table '{table}' does not exist.")

        records_to_delete = [
            record_id for record_id, record in self.data[table].items()
            if all(record.get(key) == value for key, value in query.items())
        ]

        if not records_to_delete:
            raise ValueError(f"No records found matching query: {query}")

        record_id = records_to_delete[0]
        record = self.data[table][record_id]
        _m_id = record["_m_id"]
        await self._update_index_for_record(table, record, record_id, operation='remove')
        del self.data[table][record_id]
        return _m_id

    async def delete_table(self, table):
        if table not in self.data:
            raise ValueError(f"Table '{table}' does not exist.")

        for record_id, record in self.data[table].items():
            await self._update_index_for_record(table, record, record_id, operation='remove')

        del self.data[table]
        del self.schemas[table]

    async def fetch_all(self, table=None):
        if table:
            return self.data[table]
        return self.data




from setuptools import setup

requirements = ['aiohttp', 'requests', 'asyncio']
readme = ''
with open("README.md", encoding="utf-8") as f:
    readme = f.read()
setup(
    name='GramDB',
    author='ishikki-Akabane',
    author_email='ishikkiakabane@outlook.com',
    version='1.0.0',
    long_description=readme,
    long_description_content_type="text/markdown",
    url='https://github.com/ishikki-akabane/GramDB',
    license='GNU General Public License v3.0',
    classifiers=[
        "Framework :: AsyncIO",
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        "Natural Language :: English",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3.7',
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Build Tools",

    ],
    description='Unlimited Capacity Database For Programming Using Telegram as Database.',
    include_package_data=True,
    keywords=['telegram', 'db', 'database', 'free', 'storage', 'code'],
    install_requires=requirements
)
